/*
 *      Stripped down encoder frontend for lamedropXPd
 *
 *      Copyright (c) 1999 Mark Taylor
 *                    2000 Takehiro TOMIANGA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* $Id: main.c,v 1.73 2001/10/29 22:00:15 markt Exp $ */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <assert.h>
#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char   *strchr(), *strrchr();
#endif

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if defined(_WIN32)
# include <windows.h>
#endif


#include "lame.h"

#include "get_audio.h"
#include "portableio.h"
#include "encode.h"
#include "misc.h"


#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

static int
write_xing_frame(lame_global_flags * gf, FILE * outf, size_t offset)
{
    unsigned char mp3buffer[LAME_MAXMP3BUFFER];
    size_t  imp3, owrite;
    
    imp3 = lame_get_lametag_frame(gf, mp3buffer, sizeof(mp3buffer));
    if (imp3 <= 0) {
        return 0;
    }
    if (imp3 > sizeof(mp3buffer)) {
        error_handler("Error writing LAME-tag frame: buffer too small: buffer size=%d  frame size=%d\n"
                    , sizeof(mp3buffer), imp3);
        return -1;
    }
    if (fseek(outf, offset, SEEK_SET) != 0) {
        error_handler("fatal error: can't update LAME-tag frame!\n");
        return -1;
    }
    owrite = (int) fwrite(mp3buffer, 1, imp3, outf);
    if (owrite != imp3) {
        error_handler("Error writing LAME-tag \n");
        return -1;
    }
    return imp3;
}

static int
write_id3v1_tag(lame_t gf, FILE * outf)
{
    unsigned char mp3buffer[128];
    int     imp3, owrite;

    imp3 = lame_get_id3v1_tag(gf, mp3buffer, sizeof(mp3buffer));
    if (imp3 <= 0) {
        return 0;
    }
    if ((size_t) imp3 > sizeof(mp3buffer)) {
        error_handler("Error writing ID3v1 tag: buffer too small: buffer size=%d  ID3v1 size=%d\n",
                     sizeof(mp3buffer), imp3);
        return 0;       /* not critical */
    }
    owrite = (int) fwrite(mp3buffer, 1, imp3, outf);
    if (owrite != imp3) {
        error_handler("Error writing ID3v1 tag \n");
        return 1;
    }
    return 0;
}

int
lame_encoder(lame_global_flags * gf, FILE * outf, char *inPath, char *outPath, le_enc_opt *opt)
{
	unsigned char mp3buffer[LAME_MAXMP3BUFFER];
	int     Buffer[2][1152];
	int     iread, imp3, owrite;
    size_t  id3v2_size;

	int     frames = 0;

    id3v2_size = lame_get_id3v2_tag(gf, 0, 0);
    if (id3v2_size > 0) {
        unsigned char *id3v2tag = malloc(id3v2_size);
        if (id3v2tag != 0) {
            imp3 = lame_get_id3v2_tag(gf, id3v2tag, id3v2_size);
            owrite = (int) fwrite(id3v2tag, 1, imp3, outf);
            free(id3v2tag);
            if (owrite != imp3) {
                error_handler("Error writing ID3v2 tag \n");
                return 1;
            }
        }
    }
  
	/* encode until we hit eof */
	do {
		/* read in 'iread' samples */
		iread = get_audio(gf, Buffer, opt);

        if (iread >= 0) {
			frames += iread;

			/* Call progress update every 50 frames */
			if (frames % 50 == 0)
				opt->progress_update(lame_get_num_samples(gf), frames);

			/* encode */
			imp3 = lame_encode_buffer_int(gf, Buffer[0], Buffer[1], iread, mp3buffer, sizeof(mp3buffer));

			/* was our output buffer big enough? */
			if (imp3 < 0) {
				if (imp3 == -1)
					error_handler("mp3 buffer is not big enough... \n");
				else
					error_handler("mp3 internal error:  error code=%i\n", imp3);
				return 1;
			}

			owrite = (int) fwrite(mp3buffer, 1, imp3, outf);
			if (owrite != imp3) {
				error_handler("Error writing mp3 output \n");
				return 1;
			}
		}
	} while (iread > 0);

	imp3 = lame_encode_flush(gf, mp3buffer, sizeof(mp3buffer)); /* may return one more mp3 frame */

	if (imp3 < 0) {
		if (imp3 == -1)
			error_handler("mp3 buffer is not big enough... \n");
		else
			error_handler("mp3 internal error:  error code=%i\n", imp3);
		return 1;
	}

	opt->progress_update(lame_get_num_samples(gf), frames);

	owrite = (int) fwrite(mp3buffer, 1, imp3, outf);
	if (owrite != imp3) {
		error_handler("Error writing mp3 output \n");
		return 1;
	}

    imp3 = write_id3v1_tag(gf, outf);
    
    if (imp3) {
        return 1;
    }
    write_xing_frame(gf, outf, id3v2_size);

	return 0;
}
/*
 * end of encode.c
 */
