/*
 *      Sets encoding parameters, builds list of files for
 *      encoding and controls encoding process.
 *
 *	Copyright (c) 2002 John Edwards
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <windows.h>
#include <time.h>
#include <string.h>

#include "main.h"
#include "misc.h"
#include "encode.h"
#include "audio.h"
#include "lame.h"

static const char *const genre_names[] = {
    /* NOTE: The spelling of these genre names is identical to those found in
     * Winamp and mp3info.
     */
    "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge",
    "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B",
    "Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative", "Ska",
    "Death Metal", "Pranks", "Soundtrack", "Euro-Techno", "Ambient", "Trip-Hop",
    "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical", "Instrumental",
    "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise", "Alternative Rock",
    "Bass", "Soul", "Punk", "Space", "Meditative", "Instrumental Pop",
    "Instrumental Rock", "Ethnic", "Gothic", "Darkwave", "Techno-Industrial",
    "Electronic", "Pop-Folk", "Eurodance", "Dream", "Southern Rock", "Comedy",
    "Cult", "Gangsta", "Top 40", "Christian Rap", "Pop/Funk", "Jungle",
    "Native US", "Cabaret", "New Wave", "Psychedelic", "Rave",
    "Showtunes", "Trailer", "Lo-Fi", "Tribal", "Acid Punk", "Acid Jazz",
    "Polka", "Retro", "Musical", "Rock & Roll", "Hard Rock", "Folk",
    "Folk-Rock", "National Folk", "Swing", "Fast Fusion", "Bebob", "Latin",
    "Revival", "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock",
    "Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock",
    "Big Band", "Chorus", "Easy Listening", "Acoustic", "Humour", "Speech",
    "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony", "Booty Bass",
    "Primus", "Porn Groove", "Satire", "Slow Jam", "Club", "Tango", "Samba",
    "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle", "Duet",
    "Punk Rock", "Drum Solo", "Acappella", "Euro-House", "Dance Hall",
    "Goa", "Drum & Bass", "Club-House", "Hardcore", "Terror", "Indie",
    "BritPop", "Negerpunk", "Polsk Punk", "Beat", "Christian Gangsta",
    "Heavy Metal", "Black Metal", "Crossover", "Contemporary Christian",
    "Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop",
    "SynthPop"
};

#define GENRE_NAME_COUNT \
    ((int)(sizeof genre_names / sizeof (const char *const)))

#define INVALID_CHARS "<>:\"\\/|*?"

extern int        encoding_done;
extern int        encoding;
extern int        decoding;
extern int        animate;
extern double     file_complete;
extern int        numfiles;
extern int        totalfiles;
extern char       fileName[];

extern char       config_LAMEoutdir[MAX_PATH];

le_options        opt;

/* GLOBAL VARIABLES.
 * We need to clean this up
 */
int               swapbytes = 0;	/* force byte swapping   default=0*/
sound_file_format input_format;   
int               silent = 10;		/* VERY silent */

mp3data_struct mp3input_data;   	/* used by Ogg,FLAC and MP3 */

int               in_bitwidth=16;

CRITICAL_SECTION mutex;

typedef struct enclist_tag {
	char *filename;
	struct enclist_tag *next;
} enclist_t;

enclist_t *head = NULL;

typedef struct taglist_OggFLAC {
	char *tag;
	struct taglist_OggFLAC *next_tag;
} taglist_OggF;

taglist_OggF *head_tag = NULL;


DWORD encode_thread(LPVOID arg);

void encthread_init(void)
{
	DWORD thread_id;
	HANDLE thand;

	numfiles = 0;
	totalfiles = 0;
	file_complete = 0.0;

	InitializeCriticalSection(&mutex);

	thand = (HANDLE)CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)encode_thread, NULL, 0, &thread_id);
	if (thand == NULL) {
		/* something bad happened, might want to deal with that, maybe...
		 */
		error_handler("Failed to create encoding thread!\n");
	}
}

void encthread_addfile(char *file)
{
	char *filename;
	enclist_t *entry, *node;

	if (file == NULL) return;

	/* create entry
	 */
	filename = strdup(file);
	entry = (enclist_t *)malloc(sizeof(enclist_t));

	entry->filename = filename;
	entry->next = NULL;

	EnterCriticalSection(&mutex);

	/* insert entry
	 */
	if (head == NULL) {
		head = entry;
	} else {
		node = head;
		while (node->next != NULL)
			node = node->next;

		node->next = entry;
	}
	numfiles++;
	totalfiles++;

	LeaveCriticalSection(&mutex);
}

/* The caller is responsible for deleting the pointer
 */
char *_getfile()
{
	char *filename;
	enclist_t *entry;

	EnterCriticalSection(&mutex);

	if (head == NULL) {
		LeaveCriticalSection(&mutex);
		return NULL;
	}

	/* pop entry
	 */
	entry = head;
	head = head->next;

	filename = entry->filename;
	free(entry);

	LeaveCriticalSection(&mutex);

	return filename;
}

/* Create TAG list for using Vorbis Tags
 */
void add_TAG(char *vorbis_tag)
{
	char *tag_name;
	taglist_OggF *tag_entry, *tag_node;

	if (vorbis_tag == NULL)
		return;

	/* create entry
	 */
	utf8_decode(vorbis_tag, &tag_name);
	tag_entry = (taglist_OggF *)malloc(sizeof(taglist_OggF));

	tag_entry->tag = tag_name;
	tag_entry->next_tag = NULL;

	/* insert entry
	 */
	if (head_tag == NULL)
		head_tag = tag_entry;
	else {
		tag_node = head_tag;
		while (tag_node->next_tag != NULL)
			tag_node = tag_node->next_tag;

		tag_node->next_tag = tag_entry;
	}
}

/* The caller is responsible for deleting the pointer
 */
char *_get_tag()
{
	char *tag_name;
	taglist_OggF *tag_entry;

	if (head_tag == NULL) {
		return NULL;
	}

	/* pop entry
	 */
	tag_entry = head_tag;
	head_tag = head_tag->next_tag;

	tag_name = tag_entry->tag;
	free(tag_entry);

	return tag_name;
}

FILE   *
init_files(lame_global_flags * gf, char *inPath, char *outPath, int *enc_delay, int *enc_padding, le_enc_opt *enc_opts)
{
	FILE   *outf;
	/* Mostly it is not useful to use the same input and output name.
	 * This test is very easy and buggy and don't recognize different names
	 * assigning the same file
	 */
	if (0 != strcmp("-", outPath) && 0 == strcmp(inPath, outPath)) {
		error_handler("Input file and Output file are the same. Abort.\n");
		return NULL;
	}

	/* Open the wav/aiff/raw pcm or mp3 input file.  This call will
	 * open the file, try to parse the headers and
	 * set gf.samplerate, gf.num_channels, gf.num_samples.
	 * if you want to do your own file input, skip this call and set
	 * samplerate, num_channels and num_samples yourself.
	 */
	init_infile(gf, inPath, enc_delay, enc_padding, enc_opts);
	if ((outf = init_outfile(outPath, lame_get_decode_only(gf))) == NULL) {
		error_handler("Can't init outfile '%s'\n", outPath);
		return NULL;
	}

	return outf;
}

/* This is more or less the same as strncasecmp - but that doesn't exist
 * everywhere, and this is a fairly trivial function, so we include it 
 */
int tag_compare (const char *s1, const char *s2, int n)
{
	int c=0;
	while (c < n) {
		if (toupper(s1[c]) != toupper(s2[c]))
			return !0;
		c++;
	}
	return 0;
}

/* Would use real "strcasecmp" but it isn't portable
 */
static int local_strcasecmp ( const char* s1, const char* s2 )
{
	unsigned char c1;
	unsigned char c2;
    
	do {
		c1 = tolower(*s1);
		c2 = tolower(*s2);
		if (!c1) {
			break;
		}
		++s1;
		++s2;
	} while (c1 == c2);
	return c1 - c2;
}

/* LAME is a simple frontend which just uses the file extension
 * to determine the file type.  Trying to analyze the file
 * contents is well beyond the scope of LAME and should not be added.
 */
static int filename_to_type ( const char* FileName )
{
    char *ext;
    
    if (ext = strrchr(FileName, '.')) {
        if (local_strcasecmp(ext, ".mpg") == 0) return sf_mp1;
        if (local_strcasecmp(ext, ".mp1") == 0) return sf_mp1;
        if (local_strcasecmp(ext, ".mp2") == 0) return sf_mp2;
        if (local_strcasecmp(ext, ".mp3") == 0) return sf_mp3;
        if (local_strcasecmp(ext, ".wav") == 0) return sf_wave;
        if (local_strcasecmp(ext, ".aif") == 0) return sf_aiff;
#if defined(HAVE_VORBIS)
        if (local_strcasecmp(ext, ".ogg") == 0) return sf_ogg;
#endif
#if defined(HAVE_LIBFLAC)
        if (local_strcasecmp(ext, ".flac") == 0) return sf_flac;
#endif
    }
    return sf_unknown;
}

/* Set various options
 */
void encthread_setset_bitrate(int v)
{
	opt.set_bitrate = v;
}

void encthread_setmono(int v)
{
	opt.mono = v;
}

void encthread_setquality(int v)
{
	opt.quality = v;
}

void encthread_setvbr_quality(int v)
{
	opt.vbr_quality = (float)v;
}

void encthread_setset_cbr(int v)
{
	opt.set_cbr = v;
}

void encthread_setabrbitrate(int v)
{
	opt.abr_bitrate = v;
}

void encthread_setcbrbitrate(int v)
{
	opt.cbr_bitrate = v;
}

void encthread_settag_mode(int tag_mode)
{
	opt.tagMode = tag_mode;
}

void encthread_setscale_mode(int scale_mode)
{
	opt.scale_mode = scale_mode;
}

void encthread_setAUTOtagging(int tagging)
{
	opt.AUTOtag = tagging;
}

void encthread_setdelInput(int delInput)
{
	opt.delete_input = delInput;
}

void encthread_setid3(int id3tag)
{
	opt.id3Tag = id3tag;
}

void encthread_set_title(char title[64])
{
	opt.Title = strdup(title);
}

void encthread_set_artist(char artist[64])
{
	opt.Artist = strdup(artist);
}

void encthread_set_album(char album[64])
{
	opt.Album = strdup(album);
}

void encthread_set_dates(char dates[16])
{
	opt.Dates = strdup(dates);
}

void encthread_set_genre(char genre[32])
{
	opt.Genre = strdup(genre);
}

void _nothing_prog(long total, long done)
{
	/* do nothing
	 */
}

void _error(char *errormessage)
{
	/* do nothing
	 */
}

void _update(long total, long done)
{
	file_complete = (double)done / (double)total;
}

/* Processing starts here
 */
void encoding_loop(void)
{
	char 			*in_file;

	while (in_file = _getfile()) {

		char				*p;
		char				in_path[MAX_PATH], title[MAX_PATH], lpstrFilename[MAX_PATH];
		char				album[MAX_PATH], artist[MAX_PATH];
		lame_global_flags*	gfp = NULL;
		char				buf_tag1[MAX_PATH];
		char				buf_tag2[MAX_PATH];
		char				buf_tag3[MAX_PATH];
		char	            tag_buf[MAX_PATH];
		char                *oggflac_comment;
		le_enc_opt      	enc_opts;
		char 				*out_file = NULL;
		FILE 				*in = NULL, *outf = NULL;
		int 				ret;
		unsigned int		i=0, j=0, k=0;
		int					nInitReturn = 0;
		double		        gain_R = 0.0;
		double	            gain_A = 0.0;
		int 				scaling = 0;
		/* add variables for encoder delay/padding */
		int 				enc_delay = -1;
		int 				enc_padding = -1;

		/* Init the global flags structure
		 */
		gfp = lame_init();

		input_format = filename_to_type (in_file);

		/* Init the id3 tag structure
		 */
		id3tag_init(gfp);

		animate = 1;

		/* Set various encoding defaults
		 */
		enc_opts.progress_update = _update;
		enc_opts.error = _error;
		if (opt.scale_mode > 0)
			opt.tagMode = 2;
		enc_opts.scale_mode = opt.scale_mode;
		enc_opts.tag_mode = opt.tagMode;

		set_filename(in_file);

		if (out_file == NULL) {
			/* Create a filename from existing filename, replacing extension with .mp3
			 */
			unsigned int len, len_ip, len_ip_2;

			strcpy(in_path, in_file);
			len = strlen(in_path);
			while(in_path[len] != '\\') len--;
			if(in_path[len] == '\\') {
				len_ip = len + 1;
				for(i = 0; len < strlen(in_path); i++, len++)
					title[i] = in_path[len+1];
			}
			in_path[len_ip] = '\0';

			len = len_ip - 2;
			while(in_path[len] != '\\' && in_path[len] != ':') len--;
			if(in_path[len] == '\\') {
				len_ip = len - 1;
				for(i = 0; len < (strlen(in_path)-2); i++, len++)
					album[i] = in_path[len+1];
				album[i] = '\0';
			}
			else
				strcpy(album, " ");

			len_ip_2 = len_ip;
			while(in_path[len_ip] != '\\' && in_path[len_ip] != ':') len_ip--;
			if(in_path[len_ip] == '\\') {
				for(i = 0; len_ip < len_ip_2; i++, len_ip++)
					artist[i] = in_path[len_ip+1];
				artist[i] = '\0';
			}
			else
				strcpy(artist, " ");

			len = strlen(title);
			title[len] = '\0';
			while(title[len] != '.') len--;
			if(title[len] == '.')
				title[len] = '\0';

			if (config_LAMEoutdir[0]) {
				len = strlen(config_LAMEoutdir);
				if(config_LAMEoutdir[len-1] == '\\')
					config_LAMEoutdir[len-1] = '\0';
				p=config_LAMEoutdir;
			}	

			if (!config_LAMEoutdir[0])
				sprintf(lpstrFilename,"%s%s.mp3",in_path, title);
			else if (p[0]=='\\')
				sprintf(lpstrFilename,"%s%s.mp3",config_LAMEoutdir,title);
			else
				sprintf(lpstrFilename,"%s\\%s.mp3",config_LAMEoutdir,title);

			out_file = lpstrFilename;

			strcpy(fileName, out_file);
		}

		/* Set arguments to LAME encoder
		 */
		if(opt.mono) {
			lame_set_mode(gfp, MONO);
			lame_set_num_channels( gfp, 1 );
		}
		else {
			lame_set_mode(gfp, JOINT_STEREO);
			lame_set_num_channels( gfp, 2 );
		}

		lame_set_bWriteVbrTag(gfp, 1);
		if(opt.set_bitrate ==1) {
			if (opt.quality != 3)
				lame_set_quality(gfp, opt.quality);
			if(opt.set_cbr == 1) {
				lame_set_preset(gfp, opt.cbr_bitrate);
				lame_set_VBR(gfp, vbr_off);
			}
			else if(opt.set_cbr == 0) {
				lame_set_preset(gfp, opt.abr_bitrate);
			}
		}
		else {
			lame_set_VBR(gfp, vbr_mtrh);
			lame_set_VBR_quality(gfp, ((opt.vbr_quality - 1001) * -1.f) / 100);
		}

		lame_set_findReplayGain(gfp,1);
				
		/* Initialize input file.  This also sets samplerate and as much
		 * other data on the input file as is available in the headers.
		 * This also needs to be done here as the oggvorbis/FLAC tags are
		 * not available until the files have been opened.
		 */
		outf = init_files(gfp, in_file, out_file, &enc_delay, &enc_padding, &enc_opts);
		if(outf == NULL) {
			error_handler("Error initializing input and output files %s/%s", in_file, out_file);
			numfiles--;
			continue;
		}	

		/* Create and set the id3 tags
		 */
		if(opt.AUTOtag) {
			if(opt.tagMode == 1) {
				memset(buf_tag1, '\0', sizeof(buf_tag1));
				for(i = 0; i < strlen(opt.Artist); i++) {
					if(opt.Artist[i] != '%') {
						buf_tag1[strlen(buf_tag1)] = opt.Artist[i];
					}
					else if(opt.Artist[i] == '%' && opt.Artist[i+1] == '1') {
						strcat(buf_tag1, title);
						i++;
					}
					else if(opt.Artist[i] == '%' && opt.Artist[i+1] == '2') {
						strcat(buf_tag1, album);
						i++;
					}
					else if(opt.Artist[i] == '%' && opt.Artist[i+1] == '3') {
						strcat(buf_tag1, artist);
						i++;
					}
				}
				if (opt.id3Tag == 2 && strlen(buf_tag1) > 30)
					buf_tag1[30] = '\0';
				id3tag_set_artist(gfp, buf_tag1);
	
				memset(buf_tag2, '\0', sizeof(buf_tag2));
				for(i = 0; i < strlen(opt.Album); i++) {
					if(opt.Album[i] != '%') {
						buf_tag2[strlen(buf_tag2)] = opt.Album[i];
					}
					else if(opt.Album[i] == '%' && opt.Album[i+1] == '1') {
						strcat(buf_tag2, title);
						i++;
					}
					else if(opt.Album[i] == '%' && opt.Album[i+1] == '2') {
						strcat(buf_tag2, album);
						i++;
					}
					else if(opt.Album[i] == '%' && opt.Album[i+1] == '3') {
						strcat(buf_tag2, artist);
						i++;
					}
				}
				if (opt.id3Tag == 2 && strlen(buf_tag2) > 30)
					buf_tag2[30] = '\0';
				id3tag_set_album(gfp, buf_tag2);

				memset(buf_tag3, '\0', sizeof(buf_tag3));
				for(i = 0; i < strlen(opt.Title); i++) {
					if(opt.Title[i] != '%') {
						buf_tag3[strlen(buf_tag3)] = opt.Title[i];
					}
					else if(opt.Title[i] == '%' && opt.Title[i+1] == '1') {
						strcat(buf_tag3, title);
						i++;
					}	
					else if(opt.Title[i] == '%' && opt.Title[i+1] == '2') {
						strcat(buf_tag3, album);
						i++;
					}
					else if(opt.Title[i] == '%' && opt.Title[i+1] == '3') {
						strcat(buf_tag3, artist);
						i++;
					}
				if (opt.id3Tag == 2 && strlen(buf_tag3) > 30)
					buf_tag3[30] = '\0';
				}
				id3tag_set_title(gfp, buf_tag3);
				id3tag_set_year(gfp, opt.Dates);
				if(opt.Genre[0])
					id3tag_set_genre(gfp, opt.Genre);
			}
			else {
				id3tag_set_artist(gfp, artist);
				id3tag_set_album(gfp, album);
				id3tag_set_title(gfp, title);
			}
		}
		/* Use the copied the oggvorbis/FLAC tags to create and set id3 tags.
		 * If id3v1 only tagging is set, tag length is checked and truncated
		 * as appropriate. Non-standard Genre is converted to 'Other'.
		 */
		else if (opt.tagMode == 2) {
			char invalid_chars[10];
			sprintf(invalid_chars, INVALID_CHARS);
			while (oggflac_comment = _get_tag()) {
				memset(tag_buf, '\0', sizeof(tag_buf));
				if (!tag_compare(oggflac_comment, "RG_RADIO=", 9)) {
					for (j = 9, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					gain_R = atof(tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "REPLAYGAIN_TRACK_GAIN=", 22)) {
					for (j = 22, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					gain_R = atof(tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "RG_AUDIOPHILE=", 14)) {
					for (j = 14, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					gain_A = atof(tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "REPLAYGAIN_ALBUM_GAIN=", 22)) {
					for (j = 22, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					gain_A = atof(tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "ARTIST=", 7)) {
					for (j = 7, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					for (i = 0; i < 9; i++) {
						for (j = 0; j < strlen(tag_buf); j++) {
							if (tag_buf[j] == invalid_chars[i])
								tag_buf[j] = '_';
						}
					}
					if (opt.id3Tag == 2 && strlen(tag_buf) > 30)
						tag_buf[30] = '\0';
					if (tag_buf[0])
						id3tag_set_artist(gfp, tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "ALBUM=", 6)) {
					for (j = 6, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					for (i = 0; i < 9; i++) {
						for (j = 0; j < strlen(tag_buf); j++) {
							if (tag_buf[j] == invalid_chars[i])
								tag_buf[j] = '_';
						}
					}
					if (opt.id3Tag == 2 && strlen(tag_buf) > 30)
						tag_buf[30] = '\0';
					if (tag_buf[0])
						id3tag_set_album(gfp, tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "TRACKNUMBER=", 12)) {
					for (j = 12, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					for (i = 0; i < 9; i++) {
						for (j = 0; j < strlen(tag_buf); j++) {
							if (tag_buf[j] == invalid_chars[i])
								tag_buf[j] = '_';
						}
					}
					if (opt.id3Tag == 2 && strlen(tag_buf) > 3)
						tag_buf[3] = '\0';
					if (tag_buf[0])
						id3tag_set_track(gfp, tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "TITLE=", 6)) {
					for (j = 6, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					for (i = 0; i < 9; i++) {
						for (j = 0; j < strlen(tag_buf); j++) {
							if (tag_buf[j] == invalid_chars[i])
								tag_buf[j] = '_';
						}
					}
					if (opt.id3Tag == 2 && strlen(tag_buf) > 30)
						tag_buf[30] = '\0';
					if (tag_buf[0])
						id3tag_set_title(gfp, tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "DATE=", 5)) {
					for (j = 5, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					for (i = 0; i < 9; i++) {
						for (j = 0; j < strlen(tag_buf); j++) {
							if (tag_buf[j] == invalid_chars[i])
								tag_buf[j] = '_';
						}
					}
					if (opt.id3Tag == 2 && strlen(tag_buf) > 4)
						tag_buf[4] = '\0';
					if (tag_buf[0])
						id3tag_set_year(gfp, tag_buf);
				}
				else if (!tag_compare(oggflac_comment, "GENRE=", 6)) {
					for (j = 6, k = 0; j <= strlen(oggflac_comment); j++, k++)
						tag_buf[k] = oggflac_comment[j];
					for (i = 0; i < 9; i++) {
						for (j = 0; j < strlen(tag_buf); j++) {
							if (tag_buf[j] == invalid_chars[i])
								tag_buf[j] = '_';
						}
					}
					if (opt.id3Tag == 2 && tag_buf[0]) {
						int     i, genre_found = 0;
						for (i = 0; i < GENRE_NAME_COUNT; ++i) {
							if (!local_strcasecmp(tag_buf, genre_names[i])) {
								genre_found = 1;
								break;
							}
						}
						if (!genre_found)
							strcpy(tag_buf, "Other");
					}
					if (tag_buf[0])
						id3tag_set_genre(gfp, tag_buf);
				}
				if (oggflac_comment[0])
					free(oggflac_comment);
			}
		}

		/* Apply Scaling, if set
		 */
		if (opt.scale_mode == 1) {	// Title Gain
			opt.scale = pow(10., gain_R * 0.05);
			enc_opts.scale = opt.scale;
		}
		else if (opt.scale_mode == 2) {	// Album Gain
			opt.scale = pow(10., gain_A * 0.05);
			enc_opts.scale = opt.scale;
		}
		if (input_format != sf_ogg && opt.scale_mode > 0) {
			setup_scaler(&enc_opts, (float)opt.scale);
			scaling = 1;
		}

		/* Set id3 tag type options
		 */
		if(opt.tagMode == 1 || opt.tagMode == 2) {
			switch (opt.id3Tag) {
				case 1:
					id3tag_add_v2(gfp);
					break;
				case 2:
					id3tag_v1_only(gfp);
					break;
				case 3:
					id3tag_v2_only(gfp);
					break;
				default:
					break;
			}
		}

		/* Turn off automatic writing of ID3 tag data into mp3 stream 
		 * we have to call it before 'lame_init_params', because that
		 * function would spit out ID3v2 tag data.
		 */
		lame_set_write_id3tag_automatic(gfp, 0);

		/* Now that all the options are set, lame needs to analyze them and
		 * set some more internal options and check for problems
		 */
		if (0 != (nInitReturn = lame_init_params(gfp))) {
			error_handler("Failed to initialize lame encoder correctly.\n");
			numfiles--;
			continue;
		}

		/* Encode a single input file
		 */
		ret = lame_encoder(gfp, outf, in_file, out_file, &enc_opts);
            
		/* Clean up and close files
		 */
		if (scaling) {
			clear_scaler(&enc_opts);
			scaling = 0;
		}

//		if (out_file) free(out_file);
		fclose(outf); /* close the output file */

		if(input_format == sf_wave)
			wav_close(enc_opts.readdata);

		if(input_format == sf_flac)
			flac_close(enc_opts.readdata);

		if(input_format != sf_ogg)
			close_infile(); /* close the input file */

		if (opt.delete_input) {
			if (remove(in_file) != 0) {
				error_handler("Error deleting input file '%s'\n", in_file);
				continue;
			}
		}

		numfiles--;
		lame_close(gfp);

	} /* Finished this file, loop around to next... */
}

void decoding_loop(void)
{
	char 			*in_file;

	while (in_file = _getfile()) {
		lame_global_flags*	gfp = NULL;
		char            	in_path[MAX_PATH], title[MAX_PATH], outfileName[MAX_PATH];
		le_enc_opt      	enc_opts;
		char 				*out_file = NULL;
		FILE 				*in = NULL, *outf = NULL;
		int 				j=0, ret;
		int					nInitReturn = 0;
		/* add variables for encoder delay/padding */
		int 				enc_delay = -1;
		int 				enc_padding = -1;

		/* Init the global flags structure
		 */
		gfp = lame_init();

		(void)lame_set_decode_only(gfp, 1);
		input_format = filename_to_type (in_file);

		animate = 1;

		/* Set various encoding defaults
		 */
		enc_opts.progress_update = _update;
		enc_opts.error = _error;

		set_filename(in_file);

		if (out_file == NULL) {
			/* Create a filename from existing filename, replacing extension with .mp3
			 */
			char *p;
			unsigned int i, len, len_ip, len_ip_2;

			strcpy(in_path, in_file);
			len = strlen(in_path);
			while(in_path[len] != '\\') len--;
			if(in_path[len] == '\\') {
				len_ip = len + 1;
				for(i = 0; len < strlen(in_path); i++, len++)
					title[i] = in_path[len+1];
			}
			in_path[len_ip] = '\0';
			len = strlen(title);
			title[len] = '\0';
			while(title[len] != '.') len--;
			if(title[len] == '.')
				title[len] = '\0';

			if (config_LAMEoutdir[0]) {
				len = strlen(config_LAMEoutdir);
				if(config_LAMEoutdir[len-1] == '\\')
					config_LAMEoutdir[len-1] = '\0';
				p=config_LAMEoutdir;
			}	

			if (!config_LAMEoutdir[0])
				sprintf(outfileName,"%s%s.wav",in_path, title);
			else if (p[0]=='\\')
				sprintf(outfileName,"%s%s.wav",config_LAMEoutdir,title);
			else
				sprintf(outfileName,"%s\\%s.wav",config_LAMEoutdir,title);

			out_file = outfileName;

			strcpy(fileName, out_file);
		}

		/* Initialize input file.  This also sets samplerate and as much
		 * other data on the input file as is available in the headers.
		 */

		outf = init_files(gfp, in_file, out_file, &enc_delay, &enc_padding, &enc_opts);
		if(outf == NULL) {
			error_handler("Error initializing input and output files %s/%s", in_file, out_file);
			numfiles--;
			continue;
		}	

		/* Now that all the options are set, lame needs to analyze them and
		 * set some more internal options and check for problems
		 */
		if (0 != (nInitReturn = lame_init_params(gfp))) {
			error_handler("Failed to initialize lame decoder correctly.\n");
			numfiles--;
			continue;
		}


		/* decode a single input file
		 */
		ret = lame_decoder(gfp, outf, 0, in_file, out_file, &enc_opts, &enc_delay, &enc_padding);

		/* Clean up and close files
		 */
		fclose(outf); /* close the output file */
		close_infile(); /* close the input file */
		lame_close(gfp);

		numfiles--;
	} /* Finished this file, loop around to next... */
}

DWORD encode_thread(LPVOID arg)
{
	while (!encoding_done){

		if(encoding) {
			encoding_loop();
		}
		else if(decoding) {
			decoding_loop();
		}
		encoding = decoding = 0;
		file_complete = 0.0;
		animate = 0;
		totalfiles = 0;
		numfiles = 0;

		Sleep(500);
	} 

	DeleteCriticalSection(&mutex);
	return (0);
}
/* end of encthread.c
 */
