/*
 *	Get Audio routines source file
 *
 *	Copyright (c) 1999 Albert L Faber
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* $Id: get_audio.c,v 1.121 2007/11/01 19:52:43 robert Exp $ */


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <assert.h>

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char   *strchr(), *strrchr();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif


#define         MAX_U_32_NUM            0xFFFFFFFF


#include <math.h>
#include <sys/stat.h>

#ifdef __sun__
/* woraround for SunOS 4.x, it has SEEK_* defined here */
#include <unistd.h>
#endif

#include "lame.h"
#include "main.h"
#include "get_audio.h"
#include "audio.h"
#include "portableio.h"
#include "timestatus.h"
#include "misc.h"
#include "lametime.h"
//#include "console.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

extern sound_file_format input_format;   

/* global data for get_audio.c. */
typedef struct get_audio_global_data {
    int     count_samples_carefully;
    int     pcmbitwidth;
    int     pcmswapbytes;
    int     pcm_is_unsigned_8bit;
    unsigned int num_samples_read;
    FILE   *musicin;
} get_audio_global_data;

static get_audio_global_data global = { 0, 0, 0, 0, 0, 0 };



int     lame_decode_initfile(FILE * fd, mp3data_struct * mp3data, int *enc_delay, int *enc_padding);

/* read mp3 file until mpglib returns one frame of PCM data */
int     lame_decode_fromfile(FILE * fd, short int pcm_l[], short int pcm_r[],
                             mp3data_struct * mp3data);

/* and for Vorbis: */
int     lame_decode_ogg_initfile( lame_global_flags*  gfp,
                                  FILE*               fd,
                                  le_enc_opt*         opt,
                                  mp3data_struct*     mp3data );
int     lame_decode_ogg_fromfile( lame_global_flags*  gfc,
                                  FILE*               fd,
                                  short int           pcm_l[],
                                  short int           pcm_r[],
                                  le_enc_opt*         opt,
                                  mp3data_struct*     mp3data );

static int read_samples_pcm(lame_global_flags * const gfp, FILE * const musicin,
                            short int mpg123pcm[2][1152], le_enc_opt *enc_opts);
static int read_samples_mp3(lame_global_flags * const gfp, FILE * const musicin,
                            short int mpg123pcm[2][1152]);
static int read_samples_ogg(lame_global_flags * const gfp, FILE * const musicin,
                            short int mpg123pcm[2][1152], le_enc_opt *enc_opts);
void    CloseSndFile(sound_file_format input, FILE * musicin);
FILE   *OpenSndFile(lame_global_flags * gfp, char *, int *enc_delay, int *enc_padding, le_enc_opt *enc_opts);


static size_t
min_size_t(size_t a, size_t b)
{
    if (a < b) {
        return a;
    }
    return b;
}

enum ByteOrder machine_byte_order(void);

enum ByteOrder
machine_byte_order(void)
{
    long one= 1;
    return !(*((char *)(&one))) ? ByteOrderBigEndian : ByteOrderLittleEndian;
}



/* Replacement for forward fseek(,,SEEK_CUR), because fseek() fails on pipes */


static int
fskip(FILE * fp, long offset, int whence)
{
#ifndef PIPE_BUF
    char    buffer[4096];
#else
    char    buffer[PIPE_BUF];
#endif

/* S_ISFIFO macro is defined on newer Linuxes */
#ifndef S_ISFIFO
# ifdef _S_IFIFO
    /* _S_IFIFO is defined on Win32 and Cygwin */
#  define	S_ISFIFO(m)	(((m)&_S_IFIFO) == _S_IFIFO)
# endif
#endif

#ifdef S_ISFIFO
    /* fseek is known to fail on pipes with several C-Library implementations
       workaround: 1) test for pipe
       2) for pipes, only relatvie seeking is possible
       3)            and only in forward direction!
       else fallback to old code
     */
    {
        int const fd = fileno(fp);
        struct stat file_stat;

        if (fstat(fd, &file_stat) == 0) {
            if (S_ISFIFO(file_stat.st_mode)) {
                if (whence != SEEK_CUR || offset < 0) {
                    return -1;
                }
                while (offset > 0) {
                    size_t const bytes_to_skip = min_size_t(sizeof(buffer), offset);
                    size_t const read = fread(buffer, 1, bytes_to_skip, fp);
                    if (read < 1) {
                        return -1;
                    }
                    offset -= read;
                }
                return 0;
            }
        }
    }
#endif
    if (0 == fseek(fp, offset, whence)) {
        return 0;
    }

    if (whence != SEEK_CUR || offset < 0) {
        if (silent < 10) {
            error_handler
                ("fskip problem: Mostly the return status of functions is not evaluate so it is more secure to polute <stderr>.\n");
        }
        return -1;
    }

    while (offset > 0) {
        size_t const bytes_to_skip = min_size_t(sizeof(buffer), offset);
        size_t const read = fread(buffer, 1, bytes_to_skip, fp);
        if (read < 1) {
            return -1;
        }
        offset -= read;
    }

    return 0;
}


FILE   *
init_outfile(char *outPath, int decode)
{
    FILE   *outf;
#ifdef __riscos__
    char   *p;
#endif

    /* open the output file */
    if (0 == strcmp(outPath, "-")) {
        lame_set_stream_binary_mode(outf = stdout);
    }
    else {
        if ((outf = fopen(outPath, "w+b")) == NULL)
            return NULL;
#ifdef __riscos__
        /* Assign correct file type */
        for (p = outPath; *p; p++) /* ugly, ugly to modify a string */
            switch (*p) {
            case '.':
                *p = '/';
                break;
            case '/':
                *p = '.';
                break;
            }
        SetFiletype(outPath, decode ? 0xFB1 /*WAV*/ : 0x1AD /*AMPEG*/);
#else
        (void) decode;
#endif
    }
    return outf;
}






void
init_infile(lame_global_flags * gfp, char *inPath, int *enc_delay, int *enc_padding, le_enc_opt *enc_opts)
{
    /* open the input file */
    global.count_samples_carefully = 0;
    global.num_samples_read = 0;
    global.pcmbitwidth = in_bitwidth;
    global.pcmswapbytes = swapbytes;
    global.pcm_is_unsigned_8bit = 0;
    global.musicin = OpenSndFile(gfp, inPath, enc_delay, enc_padding, enc_opts);
}

void
close_infile(void)
{
    CloseSndFile(input_format, global.musicin);
}

static int
get_audio_common(lame_global_flags * const gfp,
                 int buffer[2][1152], short buffer16[2][1152], le_enc_opt *enc_opts);

/************************************************************************
*
* get_audio()
*
* PURPOSE:  reads a frame of audio data from a file to the buffer,
*   aligns the data for future processing, and separates the
*   left and right channels
*
************************************************************************/
int
get_audio(lame_global_flags * const gfp, int buffer[2][1152], le_enc_opt *enc_opts)
{
    return (get_audio_common(gfp, buffer, NULL, enc_opts));
}

/*
  get_audio16 - behave as the original get_audio function, with a limited
                16 bit per sample output
*/
int
get_audio16(lame_global_flags * const gfp, short buffer[2][1152], le_enc_opt *enc_opts)
{
    return (get_audio_common(gfp, NULL, buffer, enc_opts));
}

/************************************************************************
  get_audio_common - central functionality of get_audio*
    in: gfp
        buffer    output to the int buffer or 16-bit buffer
   out: buffer    int output    (if buffer != NULL)
        buffer16  16-bit output (if buffer == NULL) 
returns: samples read
note: either buffer or buffer16 must be allocated upon call
*/
static int
get_audio_common(lame_global_flags * const gfp, int buffer[2][1152], short buffer16[2][1152], le_enc_opt *enc_opts)
{
    int     num_channels = lame_get_num_channels(gfp);
    int     insamp[2 * 1152];
    short   buf_tmp16[2][1152];
    int     samples_read;
    int     framesize;
    int     samples_to_read;
    unsigned int remaining, tmp_num_samples;
    int     i;
    int    *p;

    /* 
     * NOTE: LAME can now handle arbritray size input data packets,
     * so there is no reason to read the input data in chuncks of
     * size "framesize".  EXCEPT:  the LAME graphical frame analyzer 
     * will get out of sync if we read more than framesize worth of data.
     */

    samples_to_read = framesize = lame_get_framesize(gfp);
    assert(framesize <= 1152);

    /* get num_samples */
    tmp_num_samples = lame_get_num_samples(gfp);

    /* if this flag has been set, then we are carefull to read
     * exactly num_samples and no more.  This is useful for .wav and .aiff
     * files which have id3 or other tags at the end.  Note that if you
     * are using LIBSNDFILE, this is not necessary 
     */
    if (global.count_samples_carefully) {
        remaining = tmp_num_samples - Min(tmp_num_samples, global.num_samples_read);
        if (remaining < (unsigned int) framesize && 0 != tmp_num_samples)
            /* in case the input is a FIFO (at least it's reproducible with
               a FIFO) tmp_num_samples may be 0 and therefore remaining
               would be 0, but we need to read some samples, so don't
               change samples_to_read to the wrong value in this case */
            samples_to_read = remaining;
    }

    switch (input_format) {
    case sf_mp1:
    case sf_mp2:
    case sf_mp3:
	if( buffer != NULL )
	    samples_read = read_samples_mp3( gfp, global.musicin, buf_tmp16 );
	else
	    samples_read = read_samples_mp3( gfp, global.musicin, buffer16 );
        break;
    case sf_ogg:
	if( buffer != NULL )
	    samples_read = read_samples_ogg( gfp, global.musicin, buf_tmp16, enc_opts );
	else
	    samples_read = read_samples_ogg( gfp, global.musicin, buffer16, enc_opts );
        break;
    default:
	if( buffer != NULL )
	    samples_read = read_samples_pcm( gfp, global.musicin, buf_tmp16, enc_opts );
	else
	    samples_read = read_samples_pcm( gfp, global.musicin, buffer16, enc_opts );
        break;
    }

    if (buffer != NULL) {
        for (i = samples_read; --i >= 0;)
            buffer[0][i] = buf_tmp16[0][i] << (8 * sizeof(int) - 16);
        if (num_channels == 2) {
            for (i = samples_read; --i >= 0;)
                buffer[1][i] = buf_tmp16[1][i] << (8 * sizeof(int) - 16);
        }
        else if (num_channels == 1) {
            memset(buffer[1], 0, samples_read * sizeof(int));
        }
        else
            assert(0);
    }


    /* if num_samples = MAX_U_32_NUM, then it is considered infinitely long.
       Don't count the samples */
    if (tmp_num_samples != MAX_U_32_NUM)
        global. num_samples_read += samples_read;

    return samples_read;
}

static int
read_samples_ogg(lame_global_flags * const gfp, FILE * const musicin, short int oggpcm[2][1152], le_enc_opt *enc_opts)
{
    int     out = 0;

#ifdef HAVE_VORBIS
    static const char type_name[] = "Ogg Vorbis file";

    out = lame_decode_ogg_fromfile (gfp, musicin, oggpcm[0], oggpcm[1], enc_opts, &mp3input_data);
    /*
     * out < 0:  error, probably EOF
     * out = 0:  not possible with lame_decode_fromfile() ???
     * out > 0:  number of output samples
     */

    if (out < 0) {
        memset (oggpcm, 0, sizeof(**oggpcm) * 2 * 1152);
        return 0;
    }

    if (lame_get_num_channels(gfp) != mp3input_data.stereo)
        error_handler(
                "Error: number of channels has changed in %s - not supported\n",
                type_name);
    if ( lame_get_in_samplerate(gfp) != mp3input_data.samplerate )
        error_handler(
                "Error: sample frequency has changed in %s - not supported\n",
                type_name);

#else
    out = -1;           /* wanna read ogg without vorbis support? */
#endif

    return out;
}



static int
read_samples_mp3(lame_global_flags * const gfp, FILE * const musicin, short int mpg123pcm[2][1152])
{
    int     out;
#if defined(HAVE_MPGLIB)
    static const char type_name[] = "MP3 file";

    out = lame_decode_fromfile(musicin, mpg123pcm[0], mpg123pcm[1], &mp3input_data);
    /*
     * out < 0:  error, probably EOF
     * out = 0:  not possible with lame_decode_fromfile() ???
     * out > 0:  number of output samples
     */
    if (out < 0) {
        memset(mpg123pcm, 0, sizeof(**mpg123pcm) * 2 * 1152);
        return 0;
    }

    if (lame_get_num_channels(gfp) != mp3input_data.stereo) {
        if (silent < 10) {
            error_handler("Error: number of channels has changed in %s - not supported\n",
                         type_name);
        }
        out = -1;
    }
    if (lame_get_in_samplerate(gfp) != mp3input_data.samplerate) {
        if (silent < 10) {
            error_handler("Error: sample frequency has changed in %s - not supported\n", type_name);
        }
        out = -1;
    }
#else
    out = -1;
#endif
    return out;
}


int
WriteWaveHeader(FILE * const fp, const int pcmbytes,
                const int freq, const int channels, const int bits)
{
    int     bytes = (bits + 7) / 8;

    /* quick and dirty, but documented */
    fwrite("RIFF", 1, 4, fp); /* label */
    Write32BitsLowHigh(fp, pcmbytes + 44 - 8); /* length in bytes without header */
    fwrite("WAVEfmt ", 2, 4, fp); /* 2 labels */
    Write32BitsLowHigh(fp, 2 + 2 + 4 + 4 + 2 + 2); /* length of PCM format declaration area */
    Write16BitsLowHigh(fp, 1); /* is PCM? */
    Write16BitsLowHigh(fp, channels); /* number of channels */
    Write32BitsLowHigh(fp, freq); /* sample frequency in [Hz] */
    Write32BitsLowHigh(fp, freq * channels * bytes); /* bytes per second */
    Write16BitsLowHigh(fp, channels * bytes); /* bytes per sample time */
    Write16BitsLowHigh(fp, bits); /* bits per sample */
    fwrite("data", 1, 4, fp); /* label */
    Write32BitsLowHigh(fp, pcmbytes); /* length in bytes of raw PCM data */

    return ferror(fp) ? -1 : 0;
}





/************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 *
 * OLD ISO/LAME routines follow.  Used if you dont have LIBSNDFILE
 * or for stdin/stdout support
 *
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************/



/************************************************************************
*
* read_samples()
*
* PURPOSE:  reads the PCM samples from a file to the buffer
*
*  SEMANTICS:
* Reads #samples_read# number of shorts from #musicin# filepointer
* into #sample_buffer[]#.  Returns the number of samples read.
*
************************************************************************/

static int 
read_samples_pcm(lame_global_flags * const gfp, FILE * const musicin,
                            short int mpg123pcm[2][1152], le_enc_opt *enc_opts)
{
    float   **buffer;
    int     samples_read;
    int     i, j;
    int     channels = lame_get_num_channels(gfp);

    buffer = malloc(sizeof(float *) * channels);

    for (i = 0; i < channels; i++)
        buffer[i] = malloc(1152 * sizeof(float));

    samples_read = enc_opts->read_samples(enc_opts->readdata, buffer, 1152);

    for(i = 0; i < channels; i++)
        for(j = 0; j < samples_read; j++)
            mpg123pcm[i][j] = (short)buffer[i][j];

    for (i = 0; i < channels; i++)
        if (buffer[i]) free(buffer[i]);
    if (buffer) free(buffer);

    return samples_read;
}



void
CloseSndFile(sound_file_format input, FILE * musicin)
{
    (void) input;
    if (fclose(musicin) != 0) {
        if (silent < 10) {
            error_handler("Could not close audio input file\n");
        }
        exit(2);
    }
}





FILE   *
OpenSndFile(lame_global_flags * gfp, char *inPath, int *enc_delay, int *enc_padding, le_enc_opt *enc_opts)
{
    FILE   *musicin;
    InputFormat *format;


    /* set the defaults from info incase we cannot determine them from file */
    lame_set_num_samples(gfp, MAX_U_32_NUM);


    if (!strcmp(inPath, "-")) {
        lame_set_stream_binary_mode(musicin = stdin); /* Read from standard input. */
    }
    else {
        if ((musicin = fopen(inPath, "rb")) == NULL) {
            if (silent < 10) {
                error_handler("Could not find \"%s\".\n", inPath);
            }
            exit(1);
        }
    }

    if (input_format == sf_mp1 ||
        input_format == sf_mp2 || input_format == sf_mp3) {
#ifdef HAVE_MPGLIB
        if (-1 == lame_decode_initfile(musicin, &mp3input_data, enc_delay, enc_padding)) {
            if (silent < 10) {
                error_handler("Error reading headers in mp3 input file %s.\n", inPath);
            }
            exit(1);
        }
#endif
        if (-1 == lame_set_num_channels(gfp, mp3input_data.stereo)) {
            if (silent < 10) {
                error_handler("Unsupported number of channels: %ud\n", mp3input_data.stereo);
            }
            exit(1);
        }
        (void) lame_set_in_samplerate(gfp, mp3input_data.samplerate);
        (void) lame_set_num_samples(gfp, mp3input_data.nsamp);
    }
    else if (input_format == sf_ogg) {
#ifdef HAVE_VORBIS
        if (-1 == lame_decode_ogg_initfile(gfp, musicin, enc_opts, &mp3input_data)) {
            error_handler("Error reading headers in ogg input file %s.\n", inPath);
            exit(1);
        }
        if (-1 == lame_set_num_channels(gfp, mp3input_data.stereo)) {
            error_handler("Unsupported number of channels: %ud\n", mp3input_data.stereo);
            exit( 1 );
        }
        (void) lame_set_in_samplerate(gfp, mp3input_data.samplerate);
        (void) lame_set_num_samples(gfp, mp3input_data.nsamp);
#else
        if (silent < 10) {
            error_handler("LAME not compiled with libvorbis support.\n");
        }
        exit(1);
#endif
    }
    else if (input_format == sf_raw) {
        /* assume raw PCM */
        if (silent < 10) {
            error_handler("Assuming raw pcm input file");
            if (swapbytes)
                error_handler(" : Forcing byte-swapping\n");
//          else
//              error_handler("\n");
        }
        global. pcmswapbytes = swapbytes;
    }
    else if (input_format == sf_wave || input_format == sf_flac || input_format == sf_aiff) {
        format = open_audio_file(musicin, &mp3input_data, enc_opts);
        if (!format) {
            fclose(musicin);
            /* error reported by reader */
            error_handler("Unrecognised format\n");
            exit(1);
        }
        else if (-1 == lame_set_num_channels(gfp, mp3input_data.stereo)) {
            error_handler("Unsupported number of channels: %ud\n", mp3input_data.stereo);
            exit(1);
        }

        (void) lame_set_in_samplerate(gfp, mp3input_data.samplerate);
        (void) lame_set_num_samples(gfp, mp3input_data.nsamp);
    }
    if (input_format == sf_unknown) {
        exit(1);
    }

    enc_opts->channels = mp3input_data.stereo;

    if (lame_get_num_samples(gfp) == MAX_U_32_NUM && musicin != stdin) {

        double  flen = lame_get_file_size(inPath); /* try to figure out num_samples */
        if (flen >= 0) {
            /* try file size, assume 2 bytes per sample */
            if (is_mpeg_file_format(input_format)) {
                if (mp3input_data.bitrate > 0) {
                    double  totalseconds = (flen * 8.0 / (1000.0 * mp3input_data.bitrate));
                    unsigned long tmp_num_samples =
                        (unsigned long) (totalseconds * lame_get_in_samplerate(gfp));

                    (void) lame_set_num_samples(gfp, tmp_num_samples);
                    mp3input_data.nsamp = tmp_num_samples;
                }
            }
            else {
                (void) lame_set_num_samples(gfp,
                                            (unsigned long) (flen /
                                                             (2 * lame_get_num_channels(gfp))));
            }
        }
    }
    return musicin;
}





#if defined(HAVE_MPGLIB)
static int
check_aid(const unsigned char *header)
{
    return 0 == memcmp(header, "AiD\1", 4);
}

/*
 * Please check this and don't kill me if there's a bug
 * This is a (nearly?) complete header analysis for a MPEG-1/2/2.5 Layer I, II or III
 * data stream
 */

static int
is_syncword_mp123(const void *const headerptr)
{
    const unsigned char *const p = headerptr;
    static const char abl2[16] = { 0, 7, 7, 7, 0, 7, 0, 0, 0, 0, 0, 8, 8, 8, 8, 8 };

    if ((p[0] & 0xFF) != 0xFF)
        return 0;       /* first 8 bits must be '1' */
    if ((p[1] & 0xE0) != 0xE0)
        return 0;       /* next 3 bits are also */
    if ((p[1] & 0x18) == 0x08)
        return 0;       /* no MPEG-1, -2 or -2.5 */
    switch (p[1] & 0x06) {
    default:
    case 0x00:         /* illegal Layer */
        return 0;

    case 0x02:         /* Layer3 */
        if (input_format != sf_mp3 && input_format != sf_mp123) {
            return 0;
        }
        input_format = sf_mp3;
        break;

    case 0x04:         /* Layer2 */
        if (input_format != sf_mp2 && input_format != sf_mp123) {
            return 0;
        }
        input_format = sf_mp2;
        break;

    case 0x06:         /* Layer1 */
        if (input_format != sf_mp1 && input_format != sf_mp123) {
            return 0;
        }
        input_format = sf_mp1;
        break;
    }
    if ((p[1] & 0x06) == 0x00)
        return 0;       /* no Layer I, II and III */
    if ((p[2] & 0xF0) == 0xF0)
        return 0;       /* bad bitrate */
    if ((p[2] & 0x0C) == 0x0C)
        return 0;       /* no sample frequency with (32,44.1,48)/(1,2,4)     */
    if ((p[1] & 0x18) == 0x18 && (p[1] & 0x06) == 0x04 && abl2[p[2] >> 4] & (1 << (p[3] >> 6)))
        return 0;
    if ((p[3] & 3) == 2)
        return 0;       /* reserved enphasis mode */
    return 1;
}

int
lame_decode_initfile(FILE * fd, mp3data_struct * mp3data, int *enc_delay, int *enc_padding)
{
    /*  VBRTAGDATA pTagData; */
    /* int xing_header,len2,num_frames; */
    unsigned char buf[100];
    int     ret;
    size_t  len;
    int     aid_header;
    short int pcm_l[1152], pcm_r[1152];
    int     freeformat = 0;

    memset(mp3data, 0, sizeof(mp3data_struct));
    lame_decode_init();

    len = 4;
    if (fread(buf, 1, len, fd) != len)
        return -1;      /* failed */
    if (buf[0] == 'I' && buf[1] == 'D' && buf[2] == '3') {
        if (silent < 10) {
            error_handler("ID3v2 found. "
                           "Be aware that the ID3 tag is currently lost when transcoding.\n");
        }
        len = 6;
        if (fread(&buf, 1, len, fd) != len)
            return -1;  /* failed */
        buf[2] &= 127;
        buf[3] &= 127;
        buf[4] &= 127;
        buf[5] &= 127;
        len = (((((buf[2] << 7) + buf[3]) << 7) + buf[4]) << 7) + buf[5];
        fskip(fd, len, SEEK_CUR);
        len = 4;
        if (fread(&buf, 1, len, fd) != len)
            return -1;  /* failed */
    }
    aid_header = check_aid(buf);
    if (aid_header) {
        if (fread(&buf, 1, 2, fd) != 2)
            return -1;  /* failed */
        aid_header = (unsigned char) buf[0] + 256 * (unsigned char) buf[1];
        if (silent < 10) {
            error_handler("Album ID found.  length=%i \n", aid_header);
        }
        /* skip rest of AID, except for 6 bytes we have already read */
        fskip(fd, aid_header - 6, SEEK_CUR);

        /* read 4 more bytes to set up buffer for MP3 header check */
        if (fread(&buf, 1, len, fd) != len)
            return -1;  /* failed */
    }
    len = 4;
    while (!is_syncword_mp123(buf)) {
        unsigned int i;
        for (i = 0; i < len - 1; i++)
            buf[i] = buf[i + 1];
        if (fread(buf + len - 1, 1, 1, fd) != 1)
            return -1;  /* failed */
    }

    if ((buf[2] & 0xf0) == 0) {
        if (silent < 10) {
            error_handler("Input file is freeformat.\n");
        }
        freeformat = 1;
    }
    /* now parse the current buffer looking for MP3 headers.    */
    /* (as of 11/00: mpglib modified so that for the first frame where  */
    /* headers are parsed, no data will be decoded.   */
    /* However, for freeformat, we need to decode an entire frame, */
    /* so mp3data->bitrate will be 0 until we have decoded the first */
    /* frame.  Cannot decode first frame here because we are not */
    /* yet prepared to handle the output. */
    ret = lame_decode1_headersB(buf, len, pcm_l, pcm_r, mp3data, enc_delay, enc_padding);
    if (-1 == ret)
        return -1;

    /* repeat until we decode a valid mp3 header.  */
    while (!mp3data->header_parsed) {
        len = fread(buf, 1, sizeof(buf), fd);
        if (len != sizeof(buf))
            return -1;
        ret = lame_decode1_headersB(buf, len, pcm_l, pcm_r, mp3data, enc_delay, enc_padding);
        if (-1 == ret)
            return -1;
    }

    if (mp3data->bitrate == 0 && !freeformat) {
        if (silent < 10) {
            error_handler("fail to sync...\n");
        }
        return lame_decode_initfile(fd, mp3data, enc_delay, enc_padding);
    }

    if (mp3data->totalframes > 0) {
        /* mpglib found a Xing VBR header and computed nsamp & totalframes */
    }
    else {
        /* set as unknown.  Later, we will take a guess based on file size
         * ant bitrate */
        mp3data->nsamp = MAX_U_32_NUM;
    }


    /*
       report_printf("ret = %i NEED_MORE=%i \n",ret,MP3_NEED_MORE);
       report_printf("stereo = %i \n",mp.fr.stereo);
       report_printf("samp = %i  \n",freqs[mp.fr.sampling_frequency]);
       report_printf("framesize = %i  \n",framesize);
       report_printf("bitrate = %i  \n",mp3data->bitrate);
       report_printf("num frames = %ui  \n",num_frames);
       report_printf("num samp = %ui  \n",mp3data->nsamp);
       report_printf("mode     = %i  \n",mp.fr.mode);
     */

    return 0;
}

/*
For lame_decode_fromfile:  return code
  -1     error
   n     number of samples output.  either 576 or 1152 depending on MP3 file.


For lame_decode1_headers():  return code
  -1     error
   0     ok, but need more data before outputing any samples
   n     number of samples output.  either 576 or 1152 depending on MP3 file.
*/
int
lame_decode_fromfile(FILE * fd, short pcm_l[], short pcm_r[], mp3data_struct * mp3data)
{
    int     ret = 0;
    size_t  len = 0;
    unsigned char buf[1024];

    /* first see if we still have data buffered in the decoder: */
    ret = lame_decode1_headers(buf, len, pcm_l, pcm_r, mp3data);
    if (ret != 0)
        return ret;


    /* read until we get a valid output frame */
    while (1) {
        len = fread(buf, 1, 1024, fd);
        if (len == 0) {
            /* we are done reading the file, but check for buffered data */
            ret = lame_decode1_headers(buf, len, pcm_l, pcm_r, mp3data);
            if (ret <= 0) {
                lame_decode_exit(); /* release mp3decoder memory */
                return -1; /* done with file */
            }
            break;
        }

        ret = lame_decode1_headers(buf, len, pcm_l, pcm_r, mp3data);
        if (ret == -1) {
            lame_decode_exit(); /* release mp3decoder memory */
            return -1;
        }
        if (ret > 0)
            break;
    }
    return ret;
}
#endif /* defined(HAVE_MPGLIB) */


int is_mpeg_file_format( int input_file_format )
{
    switch (input_file_format) {
    case sf_mp1:
        return 1;
    case sf_mp2:
        return 2;
    case sf_mp3:
        return 3;
    case sf_mp123:
        return -1;
    default:
        break;
    }
    return 0;
}

/* end of get_audio.c */
