/*
 *      Miscellaneous functions that don't fit elsewhere!
 *
 *	Copyright (c) 2002 John Edwards
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "misc.h"


static char *_filename;
void (*error_handler)(const char *fmt, ...) = error_dialog;



/**
 * Set the current input file name.
 */
void set_filename(char *filename)
{
	_filename = filename;
}

int local_strcasecmp(const char* s1, const char* s2)
{
	unsigned char c1;
	unsigned char c2;
    
	do {
		c1 = tolower(*s1);
		c2 = tolower(*s2);
		if (!c1) {
			break;
		}
		++s1;
		++s2;
	} while (c1 == c2);
	return c1 - c2;
}

/**
 * Display an error dialog, possibly adding system error information.
 */
void error_dialog(const char *fmt, ...)
{
	va_list ap;
	char msgbuf[1024];
	char *bufp = msgbuf;

	/* A really rough sanity check to protect against blatant buffer overrun */
	if (strlen(fmt) > 750) {
		sprintf(msgbuf, "%s %s", "<buffer overflow> ", fmt);
	} 
	else {
		if (_filename != NULL && strlen(_filename) < 255) {
			sprintf(msgbuf, "%s: ", _filename);
			bufp += strlen(msgbuf);
		}

		va_start(ap, fmt);
		
		vsprintf(bufp, fmt, ap);

		va_end(ap);

		if (errno != 0) {
			bufp = msgbuf + strlen(msgbuf);
			sprintf(bufp, " error is %s (%d)", strerror(errno), errno);
			errno = 0;
		}
	}

	MessageBox(NULL, msgbuf, "Error", 0);
}

void log_error(const char *fmt, ...)
{
	va_list ap;
	FILE *fp;
	char msgbuf[1024];
	char *bufp = msgbuf;

	/* A really rough sanity check to protect against blatant buffer overrun */
	if (strlen(fmt) > 750)
		sprintf(msgbuf, "%s %s", "<buffer overflow> ", fmt);
	else {
		if (_filename != NULL && strlen(_filename) < 255) {
			sprintf(msgbuf, "%s : ", _filename);
			bufp += strlen(msgbuf);
		}

		va_start(ap, fmt);

		vsprintf(bufp, fmt, ap);

		va_end(ap);

		if (errno != 0) {
			bufp = msgbuf + strlen(msgbuf);
			sprintf(bufp, " error is: %s (%d)", strerror(errno), errno);
			errno = 0;
		}
	}

	va_start(ap, fmt);

	if ((fp = fopen("lamedropXPd.log", "a")) == (FILE *)NULL)
		return;

	fprintf(fp, "%s\n", msgbuf);
	fflush(fp);
	fclose(fp);

	va_end(ap);
}

void set_use_dialogs(int use_dialogs)
{
	if (!use_dialogs)
		error_handler = error_dialog;
	else
		error_handler = log_error;
}
/*
 * end of misc.c
 */

