/* $Id: vorbis_interface.c,v 1.30 2004/01/28:13:01 jfedwards $ */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

/* LAME interface to libvorbis */

#ifdef HAVE_VORBIS
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include "vorbis/codec.h"
#include "vorbis/vorbisenc.h"
#include "vorbis/vorbisfile.h"
#include "vorbis/os.h"          // included for the vorbis ftoi routines

#include "lame.h"
#include "util.h"
#include "main.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

//#define FLOAT_SCALE (1.0f/(1<<15))
#define FLOAT_SCALE (1.0f/(1<<31))
int 			convsize = 1152;

sound_file_format       input_format;   
OggVorbis_File 		vf;
vorbis_info 		*vi;



int lame_decode_ogg_initfile( lame_global_flags*  gfp,
                              FILE*               fd,
                              le_enc_opt*         opt,
                              mp3data_struct*     mp3data )
{

//	lame_internal_flags *gfc = gfp->internal_flags;


	if(ov_open(fd, &vf, NULL, 0) < 0)
	{
//		ERRORF(gfc,"\n Input file does not appear to be an Ogg bitstream.\n");
		return -1;
	}

	vi=ov_info(&vf,-1);

	mp3data->stereo = vi->channels;
	mp3data->samplerate = vi->rate;
	mp3data->bitrate = (int)(ov_bitrate(&vf, -1)/1000);
	mp3data->nsamp=(long)ov_pcm_total(&vf,-1);
	lame_set_num_samples(gfp, (long)ov_pcm_total(&vf,-1));

	if (opt->tag_mode == 2)
	{

		char **ptr=ov_comment(&vf,-1)->user_comments;

		while(*ptr)
		{
			add_TAG(*ptr);
			++ptr;
		}
	}
	return 0;
}



/*
  For lame_decode_fromfile:  return code
  -1     error, or eof
  0     ok, but need more data before outputing any samples
  n     number of samples output.  
*/
int lame_decode_ogg_fromfile( lame_global_flags*  gfp,
                              FILE*               fd,
                              short int           pcm_l[],
                              short int           pcm_r[],
                              le_enc_opt*         opt,
                              mp3data_struct*     mp3data )
{
//	lame_internal_flags *gfc = gfp->internal_flags;
	int i,j,eof=0,eos=0;
	int current_section;
	int bout;
	float **pcm;
       
	long ret = ov_read_float(&vf, &pcm, convsize, &current_section);

	if (ret == 0) 
	{
		/* EOF */
		eof = 1;
	}
	else
	{
		if (ret < 0)
		{
			/* error in the stream.  Not a problem, just reporting it in
			 * case we (the app) cares.  In this case, we don't. 
			 */
		}
		else
		{
			/* we don't bother dealing with sample rate changes, etc, but
			 * you'll have to
			 */
			vorbis_fpu_control fpu;
			vorbis_fpu_setround(&fpu);
			bout=(ret<convsize?ret:convsize);

			/* convert floats to 16 bit signed ints.
			 */
			for(i = 0; i < vi->channels; i++)
			{
				float  *mono=pcm[i];
				if(i == 0)
				{
					for(j = 0; j < bout; j++)
					{
						/* might as well guard against clipping */
						if (mono[j] > 0.9999694)
							mono[j] = 0.9999694;
						else if (mono[j] < -1.0)
							mono[j] = -1.0;
						pcm_l[j] = (short)vorbis_ftoi(mono[j] * 32767.);
						if (opt->scale_mode > 0)
							pcm_l[j] *= opt->scale;
					}
				}
				else if (i==1)
				{
					for(j = 0; j < bout; j++)
					{
						/* might as well guard against clipping */
						if (mono[j] > 0.9999694)
							mono[j] = 0.9999694;
						else if (mono[j] < -1.0)
							mono[j] = -1.0;
						pcm_r[j] = (short)vorbis_ftoi(mono[j] * 32767.);
						if (opt->scale_mode > 0)
							pcm_r[j] *= opt->scale;
					}
				}
			}
			vorbis_fpu_restore(fpu);
		}
	}
  
	if (eof)
	{
		ov_clear(&vf);
		return ret;
	}
	return bout;
}





#if HAVE_VORBIS_ENCODER
#include "audio.h"
#include "resample.h"

ogg_stream_state  os2;  // take physical pages, weld into a logical stream of packets
ogg_page          og2;  // one Ogg bitstream page. Vorbis packets are inside
ogg_packet        op2;  // one raw packet of data for decode

vorbis_info       vi2;  // struct that stores all the static vorbis bitstream settings
vorbis_comment    vc2;  // struct that stores all the bitstream user comments
vorbis_dsp_state  vd2;  // central working state for the packet->PCM decoder
vorbis_block      vb2;  // local working space for packet->PCM decode

int oe_write_page(ogg_page *page, FILE *fp);

int               first_frame;

#define MAX_COMMENT_LENGTH 255

int lame_encode_ogg_init(lame_global_flags *gfp)
{
	lame_internal_flags *gfc = gfp->internal_flags;

	char comment[MAX_COMMENT_LENGTH+1];

	/* init the vorbis stream */
	vorbis_info_init(&vi2);
	/* setup the encoder parameters */
	if(gfp->ogg_ABR == -1) {
		vorbis_encode_setup_vbr(&vi2, gfc->channels_out, gfp->out_samplerate, 
					gfp->ogg_quality);
		/* Turn off management entirely (if it was turned on). */
		vorbis_encode_ctl(&vi2, OV_ECTL_RATEMANAGE_SET, NULL);
	} else
		vorbis_encode_setup_managed(&vi2, gfc->channels_out, gfp->out_samplerate, 
					-1, gfp->ogg_ABR, -1);
    
	vorbis_encode_setup_init(&vi2);

	/* add a comment */
	vorbis_comment_init(&vc2);

	switch(input_format)
	{
		case sf_mp1:
			vorbis_comment_add_tag(&vc2, "SOURCE", "mp1.");
			break;
		case sf_mp2:
			vorbis_comment_add_tag(&vc2, "SOURCE", "mp2.");
			break;
		case sf_mp3:
			vorbis_comment_add_tag(&vc2, "SOURCE", "mp3.");
			break;
		case sf_ogg:
			vorbis_comment_add_tag(&vc2, "SOURCE", "ogg.");
			break;
	}
	vorbis_comment_add_tag(&vc2, "ENCODED_USING", "L.A.M.E. libVorbis interface.");
     
	if(gfc->tag_spec.title) {
		strcpy(comment, gfc->tag_spec.title);
		vorbis_comment_add_tag(&vc2, "TITLE", comment);
	}
	if(gfc->tag_spec.artist) {
		strcpy(comment, gfc->tag_spec.artist);
		vorbis_comment_add_tag(&vc2, "ARTIST", comment);
	}
	if(gfc->tag_spec.album) {
		strcpy(comment,gfc->tag_spec.album);
		vorbis_comment_add_tag(&vc2, "ALBUM", comment);
	}
	if(gfc->tag_spec.year) {
		sprintf(comment, "%d", gfc->tag_spec.year);
		vorbis_comment_add_tag(&vc2, "DATE", comment);
	}
	if(gfc->tag_spec.comment) {
		strcpy(comment, gfc->tag_spec.comment);
		vorbis_comment_add_tag(&vc2, "DESCRIPTION", comment);
	}
	if(gfc->tag_spec.track) {
		sprintf(comment, "%d", gfc->tag_spec.track);
		vorbis_comment_add_tag(&vc2, "TRACK", comment);
	}
	/* TODO -- support for genre */

	if(gfp->in_samplerate != gfp->out_samplerate)
		setup_resampler(gfp);

	/* Set up the analysis state and auxiliary encoding storage */
	vorbis_analysis_init(&vd2,&vi2);
	vorbis_block_init(&vd2,&vb2);
  
	/* set up our packet->stream encoder */
	/* pick a random serial number; that way we can more likely build
	   chained streams just by concatenation */
	srand(time(NULL));
	ogg_stream_init(&os2,rand());
  
	/* Vorbis streams begin with three headers; the initial header (with
	   most of the codec setup parameters) which is mandated by the Ogg
	   bitstream spec.  The second header holds any comment fields.  The
	   third header holds the bitstream codebook.  We merely need to
	   make the headers, then pass them to libvorbis one at a time;
	   libvorbis handles the additional Ogg bitstream constraints */
  
	{
		ogg_packet header;
		ogg_packet header_comm;
		ogg_packet header_code;
    
		/* Build the packets */
		vorbis_analysis_headerout(&vd2,&vc2,&header,&header_comm,&header_code);

		/* And stream them out */
		ogg_stream_packetin(&os2,&header); /* automatically placed in its own page */
		ogg_stream_packetin(&os2,&header_comm);
		ogg_stream_packetin(&os2,&header_code);
    
		/* no need to write out here.  We'll get to that in the main loop */
	}
	first_frame = 1;
	return 0;
}



int lame_encode_ogg_finish(lame_global_flags*  gfp)
{
	/* clean up and exit.  vorbis_info_clear() must be called last */
	if(gfp->in_samplerate != gfp->out_samplerate)
		clear_resampler();

	ogg_stream_clear(&os2);
	vorbis_block_clear(&vb2);
	vorbis_dsp_clear(&vd2);
  	vorbis_comment_clear(&vc2);
	vorbis_info_clear(&vi2);

	/* ogg_page and ogg_packet structs always point to storage in
	   libvorbis.  They're never freed or manipulated directly */
	gfp->frameNum++;
	return 0;
}



int  lame_encode_ogg_frame (
	lame_global_flags*  gfp,
	int                 buffer_l[],
	int                 buffer_r[],
	const int           nsamples,
	FILE*               outf )
{
	lame_internal_flags *gfc = gfp->internal_flags;
	int  i;
	int  c;
	int  eos   = 0;
	int  bytes = 0;
	int  result;
	int bytes_written = 0;
	long out_samples = 0;
	float **buffer;
	float **in_buffer;

	in_buffer = malloc(sizeof(float *) * 2);
	in_buffer[0] = malloc(sizeof(float) * nsamples);
	in_buffer[1] = malloc(sizeof(float) * nsamples);

	if(first_frame)
	{
		while((result = ogg_stream_flush(&os2, &og2)))
		{
			if(!result) break;
			bytes = oe_write_page(&og2, outf);
			if(bytes != og2.header_len + og2.body_len)
			{
				ERRORF(gfc, "Failed writing header to output stream\n");
				return -6;
			}
			else
				bytes_written += bytes;
		}
	}

	first_frame = FALSE;

	/* expose the buffer to submit data */
	buffer = vorbis_analysis_buffer(&vd2, nsamples);

	/* change to float 
	 */
	for (i = 0; i < nsamples; i++) {
		in_buffer[0][i] = (float)buffer_l[i] * FLOAT_SCALE;
		in_buffer[1][i] = (float)buffer_r[i] * FLOAT_SCALE;
	}

	if (gfp->num_channels == 2 && gfc->channels_out == 1) {
		/* downmix to mono */
		for (i = 0; i < nsamples; i++)
			in_buffer[0][i] = (in_buffer[0][i] + in_buffer[1][i]) * 0.5;
	}
	if(gfp->in_samplerate != gfp->out_samplerate)
		out_samples = ov_resampler(in_buffer, buffer, nsamples);
	else if (gfp->num_channels == 2 && gfc->channels_out == 1) {
		for (i = 0; i < nsamples; i++)
			buffer[0][i] = in_buffer[0][i];
	} else {
		for (i = 0; i < nsamples; i++) {
			buffer[0][i] = in_buffer[0][i];
			buffer[1][i] = in_buffer[1][i];
		}
	}
	/* tell the library how much we actually submitted */
	if(gfp->in_samplerate != gfp->out_samplerate)
		vorbis_analysis_wrote(&vd2, out_samples);
	else
		vorbis_analysis_wrote(&vd2, nsamples);

	/* vorbis does some data preanalysis, then divvies up blocks for
	   more involved (potentially parallel) processing.  Get a single
	   block for encoding now */
	while(vorbis_analysis_blockout(&vd2, &vb2) == 1)
	{
		int result;
		/* analysis */
		vorbis_analysis(&vb2, NULL);
		vorbis_bitrate_addblock(&vb2);

		while(vorbis_bitrate_flushpacket(&vd2, &op2)) 
		{

			/* add the packet into the bitstream */
			ogg_stream_packetin(&os2, &op2);
    
			/* write out pages (if any) */
			while(!eos)
			{
				result = ogg_stream_pageout(&os2,&og2);
				if (result == 0)
					break;

				bytes = oe_write_page(&og2, outf);
				if(bytes != og2.header_len + og2.body_len)
				{
					ERRORF(gfc, "Failed writing data to output stream\n");
					return -6;
				}
				else
					bytes_written += bytes;
     
				if(ogg_page_eos(&og2))
					eos = 1;
			}
		}
	}
	for(c = 0; c < gfp->num_channels; c++)
		free(in_buffer[c]);
	free(in_buffer);
	gfp->frameNum++;
	return bytes_written;
}

int oe_write_page(ogg_page *page, FILE *fp)
{
	int written;
	written = fwrite(page->header,1,page->header_len, fp);
	written += fwrite(page->body,1,page->body_len, fp);

	return written;
}

#endif
#endif

/* end of vorbis_interface.c */

